@@ -2,7 +2,7 @@
 ##
 ## DTest - A Distributed test framework
 ##
-## Copyright (c) 2006,2007 Eric NOULARD and Frederik DEWEERDT 
+## Copyright (c) 2006-2008 Eric NOULARD, Lionel DUROYON and Frederik DEWEERDT 
 ##
 ## This library is free software; you can redistribute it and/or
 ## modify it under the terms of the GNU Lesser General Public
@@ -33,154 +33,11 @@ from StringIO import StringIO
 import logging
 import inspect
 import re
-
-class SessionHandler (object):
-    """Represents low-level DTest session handler.
-
-    A dtest session handler is responsible for giving
-    a DTester the low level mean to have a pty-like
-    access to the session and an eventual file manipulation
-    capability
-    Thus DTestSessionHandler may have several capabilities:
-         - command run capability
-         - file transfer capability
-
-    The command run capability includes:
-        - a way to update environment
-        - a way to send a string to pty input
-        - a way to recv data from pty output
-    """
-
-    class UnableToUpdateEnvironException(Exception):
-        """This DTest Session Handler is not capable to update environ"""
-        pass
-
-    class UnableToRunCommandException(Exception):
-        """This DTest Session Handler is not capable to run command"""
-        pass
-
-    class UnableToFileTransferException(Exception):
-        """This DTest Session Handler is not capable to transfer file"""
-        pass
-
-    class SessionFailureException(Exception):
-        """The DTest Session Handler encounter a failure"""
-        pass
-
-    logger = logging.getLogger("DTesterSessionHandler")
-    sh  = logging.StreamHandler()
-    fmt = logging.Formatter(fmt="## [%(threadName)s]|%(name)s::%(levelname)s:: %(message)s")
-    sh.setFormatter(fmt)
-    logger.addHandler(sh)
-
-    def __init__(self,runCommandCapable=False, fileTransferCapable=False):
-        """
-        Create a Default Session Handler with no capability.        
-        """
-        self.isRunCommandCapable     = runCommandCapable
-        self.isFileTransferCapable   = fileTransferCapable
-        self.sessionOpened           = False
-        self.environ                 = None
-        self.__lastReceive           = None
-        self.__lastSent              = None
-        self.CTRL_C  = chr(03)
-        self.ETX     = chr(03)
-        self.EOT     = chr(04)
-        self.CR      = chr(13)
-        self.LF      = chr(10)
-        self.CRLF    = self.CR + self.LF
-        self.NEWLINE = self.LF
-        self.ESC     = chr(27)
-        self.stdin   = None
-        self.stderr  = None
-        self.stdout  = None
-
-    def __getLastReceive(self):
-        return self.__LastReceive
-    
-    def __setLastReceive(self,string):
-        self.__lastReceive = string
-        # FIXME should add logger
-        #print "LastReceive = %s" % string
-
-    lastReceive = property(fget=__getLastReceive,fset=__setLastReceive)
-
-    def __getLastSent(self):
-        return self.__lastSent
-    
-    def __setLastSent(self,string):
-        self.__lastSent = string
-        # FIXME should add logger
-        #print "LastSent = %s" % string
-
-    lastSent = property(fget=__getLastSent,fset=__setLastSent)
-
-    def __isnotoverloaded(self,method):
-        methname = method.__name__
-        return (getattr(self.__class__,methname)==getattr(SessionHandler,methname))
-        
-    def openIfNotOpened(self):
-        """ Open the session if it is not already opened."""
-        if not self.sessionOpened:
-            self.open()
-
-    def open(self,*args,**kwargs):
-        """ Open the session """
-        self.sessionOpened = True
-
-    def close(self):
-        """ Close the session. """
-        if (not self.stdin==None):
-            self.stdin.close()
-        if (not self.stderr==None):
-            self.stderr.close()
-        if (not self.stdout==None):
-            self.stdout.close()    
-        self.sessionOpened = False 
-        
-    def updateEnviron(self,environ):
-        """ Update the environment """
-        self.environ = environ
-        for envvar in self.environ.iteritems():
-           envcmd = "%s=%s" % envvar
-           self.sendall(envcmd+self.NEWLINE) 
-
-    def send(self, string):
-        """ Send a string to the session """
-        self.lastSent = string
-        if (not self.stdin==None):
-            self.stdin.write(string)
-        if (self.__isnotoverloaded(SessionHandler.send)):
-            raise self.UnableToRunCommandException("Cannot send")        
-
-    def sendall(self, string):
-        self.lastSent = string
-        if (not self.stdin==None):
-            self.stdin.write(string)
-        if (self.__isnotoverloaded(SessionHandler.sendall)):
-            raise self.UnableToRunCommandException("Cannot sendall")
-
-    def recv_ready(self):
-        if (self.__isnotoverloaded(SessionHandler.recv_ready)):
-            raise self.UnableToRunCommandException("Cannot recv_ready")
-
-    def recv(self, size, buffer=None):
-        if (not self.stdout==None):
-            self.stdout.write(self.__lastReceive)
-        if (self.__isnotoverloaded(SessionHandler.recv)):
-            raise self.UnableToRunCommandException("Cannot recv")
-
-    def putFile(self, sourcePath, destinationPath):
-        if (self.__isnotoverloaded(SessionHandler.putFile)):
-            raise self.UnableToFileTransferException("Cannot putFile")
-
-    def getFile(self, sourcePath, destinationPath):
-        if (self.__isnotoverloaded(SessionHandler.getFile)):
-            raise self.UnableToFileTransferException("Cannot getFile")
-
+import session_handler
+from session_handler import SessionHandler
 
 class DTester (Thread):
-    """Represents an elementary DTest stakeholder.
+    """Represent an elementary DTest stakeholder.
     
     DTest falls into (eventually) communicating and concurrent
     processes, the DTesters, which will be runned concurrently
@@ -188,7 +45,7 @@ class DTester (Thread):
     sequential steps (either initialize or run steps) specified
     through addInitializeStep and/or addRunStep
     DTester method.
-    FIXME: describe DTESTMASTER 
+    FIXME: describe DTESTMASTER
     """
  
     class InvalidStepFunctionException(Exception):
@@ -209,6 +66,9 @@ class DTester (Thread):
     class SessionUnableToRunCommandException(Exception):
         """The DTest Session Handler cannot run command"""
 
+    class StepTimeOut(Exception):
+        """A step has timed out"""
+
 
     def __isbound(cls,function):
         return inspect.ismethod(function) and function.im_self != None
@@ -230,7 +90,7 @@ class DTester (Thread):
     logger.addHandler(sh)
        
     def __init__(self,name,session=SessionHandler(),timeout=10):
-        """ Create a DTester using the specified session handler.
+        """Creates a DTester using the specified Sessionhandler.
 
         @param name: the name of the DTester
         @type name: C{string}
@@ -246,13 +106,33 @@ class DTester (Thread):
         self.timeout            = timeout
         self.environ            = None
         self.dtestmaster        = None
-        self.nb_steps           = 0
+        self.__nbSteps          = 0
+        self.__ifThenElse_scope = False
         
         self.__expectDidTimeOut   = False
         self.__runSteps           = []
         self.__initializeSteps    = []
+        self.__lastStepStatus     = False;
+        
+    def __getNbSteps(self):
+        return self.__nbSteps    
+    
+    nbSteps=property(fget=__getNbSteps,doc='Number of steps in the DTest sequence')
+    
+    def getLastStepStatus(self):
+        return self.__lastStepStatus
+    
+    def getFutureLastStepStatus(self):
+        return self.getLastStepStatus             
 
+    def __getHasTimedOut(self):
+        return self.__expectDidTimeOut
 
+    def __setHasTimedOut(self,value):
+        self.__expectDidTimeOut = True
+        
+    hasTimedOut = property(fget=__getHasTimedOut,fset=__setHasTimedOut)
+        
     def __getName(self):
         return self.getName()
 
@@ -279,11 +159,14 @@ class DTester (Thread):
     
     def __setStdout(self,file):
         self.session.stdout = file
+        
+    def getRunSteps(self):
+        return self.__runSteps
 
     stdout = property(fget=__getStdout,fset=__setStdout)
 
     def registerForBarrier(self, barrierId):
-        """register for DTest barrier barrierId in the DTestMaster.
+        """Register for DTest barrier barrierId in the DTestMaster.
 
         This method does not need to be called explicitly.
         The DTester will generate the call as soon as a 'barrier'
@@ -294,7 +177,7 @@ class DTester (Thread):
         self.dtestmaster.registerForBarrier(self,barrierId)
         
     def barrier(self, barrierId, timeout=None):
-        """wait until every stakeholder have reached the specified barrier.
+        """Wait until every stakeholder have reached the specified barrier.
 
         If timeout is not specified the DTester instance-wide timeout
         is used. If specified then the specified timeout is used.
@@ -307,28 +190,51 @@ class DTester (Thread):
         """
         if (timeout==None):
             timeout = self.timeout
-        self.dtestmaster.barrier(self,barrierId,timeout)
+        return self.dtestmaster.barrier(self,barrierId,timeout)
 
-    def ok(self, *args, **kwargs):
-        """ok TAP method.
+    def ok(self, ok_nok, desc, **kwargs):
+        """Ok TAP (Test Anything Protocol) method"""
+        # If we have been given a callable object for 'ok_nok'
+        # this mean it should be called when OK is called
+        if (callable(ok_nok)):
+            callme = ok_nok()
+            return self.dtestmaster.ok(self, callme(), desc, **kwargs)
+        else:
+            return self.dtestmaster.ok(self, ok_nok, desc, **kwargs)
         
-        """
-        self.dtestmaster.ok(self, *args, **kwargs)
+    def ifThenElse(self,condition):
+        if (callable(condition)):
+            callme = condition()
+            return callme()
+        else:
+            return condition
 
     ## The following are command related API
     ## which are implemented using lower-level
-    ## session handler API
-    
+    ## session handler API    
     def runCommand(self, command=None, environ=None):
+        """Run a command on the session handler DTest.session"""
         self.session.openIfNotOpened()
         if (environ != None):
             self.session.updateEnviron(environ)
         if (command != None):
-            self.session.send(command+self.session.NEWLINE)        
+            return self.session.send(command+self.session.NEWLINE)    
+           
+    def abort(self):
+        if self.stdin != None:
+            self.stdin.flush()
+        if self.stdout != None:
+            self.stdout.flush()
+        if self.stderr != None:
+            self.stderr.flush()
 
     def __expectTimedOut(self,msg=None):
-        self.logger.warning("Expect timeout: %s" % msg)
-        self.__expectDidTimeOut = True
+        if (not self.silentExpect==True):
+            self.logger.warning("Expect timeout: %s" % msg)
+        self.abort()
+        self.__expectDidTimeOut = True        
+        # notify session 
+        self.session.hasTimedOut = True
 
     def __createExpectTimer(self,timeout=10,msg=""):
         self.__expectDidTimeOut = False
@@ -336,10 +242,17 @@ class DTester (Thread):
         t.setName("Timer-"+self.getName())
         return t
     
-    def expectFromCommand(self, pattern, timeout=None):        
-        pat       = re.compile(pattern)
+    def expectFromCommand(self, pattern, timeout=None, silent=None):    
+        """Wait until the expected pattern is received on the session handler DTester.session"""
+        if (type(pattern)==type(re.compile(""))):
+            pat = pattern
+        else:
+            pat = re.compile(pattern)
+            
         monitored = StringIO("")
 
+        self.silentExpect = silent
+
         # if no timeout is specified then use
         # the one of the DTester instance
         if timeout==None:
@@ -357,8 +270,12 @@ class DTester (Thread):
         while (not pat.search(monitored.getvalue())) and (not self.__expectDidTimeOut):
             self.session.recv(1,monitored)
 
+        # We will get there because time out handler did tell
+        # the session handler the test has timedout
         if self.__expectDidTimeOut:
-            self.logger.warn("Monitored = %s" % monitored.getvalue())
+            self.logger.debug("Monitored = %s" % monitored.getvalue())
+            # reset timeout state for the session
+            self.session.hasTimedOut = False
             return False
         else:
             if timeout != None:
@@ -366,10 +283,12 @@ class DTester (Thread):
             return True
     
     def sendToCommand(self, string):
+        """Send directly a string on the session handler DTester.session"""
         self.session.sendall(string)
         return True
     
     def waitCommandTermination(self, timeout=None):
+        """Wait until the session handler DTester.session ends"""
         if self.session.sessionOpened:
             while self.session.recv_ready():
                 self.session.recv(1)
@@ -377,19 +296,20 @@ class DTester (Thread):
 
     def waitDuring(self,duration):
         time.sleep(duration)
+        return True
     
     def terminateCommand(self):
+        """Send to the session handler DTester.session the command for terminating"""
         self.session.sendall(self.session.CTRL_C)
-        self.session.sendall(self.session.NEWLINE)
-        return True
+        return self.session.sendall(self.session.NEWLINE)        
 
     def putFile(self,srcPath,dstPath):
         self.session.openIfNotOpened()
-        self.session.putFile(srcPath,dstPath)
-
+        return self.session.putFile(srcPath,dstPath)
+        
     def getFile(self,srcPath,dstPath):
         self.session.openIfNotOpened()
-        self.session.getFile(srcPath,dstPath)
+        return self.session.getFile(srcPath,dstPath)
 
     def addInitializeStep(self, stepmethod, *args, **kwargs):
         if not callable(stepmethod):
@@ -397,7 +317,27 @@ class DTester (Thread):
         self.__initializeSteps.append((stepmethod,args,kwargs))
 
     def addRunStep(self, stepmethod, *args, **kwargs):
-
+        """Add stepmethod run step to a DTester
+        
+        @param stepmethod: the stepmethod name
+        @type stepmethod: C{string}     
+        
+        Stepmethod parameter can be :
+            
+        ok : Test Anything Protocol display method.
+        runCommand : runs a command on the associed session handler.
+        expectFromCommand : waits for the reception of a string pattern on the associated session handler. 
+        terminateCommand : terminates a command on the associated session handler.
+        barrier : indicates that a DTester has reached a barrier.
+        sendToCommand : sends a command on the associated session handler.
+        waitCommandTermination : waits the command termination on the associated session handler.
+        custom : allows to run any python "callable".
+        
+        Each stepmethod matches an implemented method of DTester, except custom which allows you
+        to define and call your own python "callable"
+        to see which args parameters must follow each stepmethod, you can look the matching method
+        """
+        
         # Handle string stepmethod case
         if isinstance(stepmethod,type("")):
             if hasattr(self,stepmethod):
@@ -405,21 +345,32 @@ class DTester (Thread):
                 stepmethod = getattr(self.__class__,stepmethod)
                 self.logger.debug("stepmethod string <%s> resolved to %r" % (stepmethod_str,stepmethod))
             else:
-                raise DTester.InvalidStepFunctionException, "Stepmethod is a unknown attribute string <%s> of class <%s>" % stepmethod, self.__class__.__name__
+                raise DTester.InvalidStepFunctionException, "Stepmethod is a unknown attribute string <%s> of class <%s>" % (stepmethod, self.__class__.__name__)
 
         # stepmethod should be callable, at least ...
         if not callable(stepmethod):
             raise DTester.StepNotCallableException, "Invalid run step:"+str(stepmethod)+"is not a callable object"
 
         if (inspect.ismethod(stepmethod)):
-            # barrier and ok method
-            # triggers special handling
+            # barrier, ok and ifThenElse
+            # method triggers special handling
             # we check for .im_func in order to capture both bounded
             # and unbounded method
             if stepmethod.im_func == DTester.barrier.im_func:
                 self.addInitializeStep(DTester.registerForBarrier,args[0])
             elif stepmethod.im_func == DTester.ok.im_func:
-                self.nb_steps += 1
+                # if an ok step is in the scope
+                # of ifThenElse then there _MUST_
+                # an ok step in each branch
+                # thus we do not count first ok step
+                if self.__ifThenElse_scope:
+                    self.__ifThenElse_scope = False
+                else:
+                    self.__nbSteps += 1
+            elif stepmethod.im_func == DTester.ifThenElse.im_func:
+                self.__ifThenElse_scope = True
+            else:
+                self.__ifThenElse_scope = False
 
         if DTester.__isfunction_or_unboundmethod(stepmethod):
             # We check first arg to be self or dtester 
@@ -450,26 +401,68 @@ class DTester (Thread):
         self.logger.info("DTester <%s> initialize complete." % self.getName())
         
     def run(self):
+        """Run the DTester thread
+        
+        This method executes each run step defined previously by addRunStep
+        """
+        
         self.logger.info("DTester <%s> begin run..." % self.getName())
+        skipNextStep   = False
+        skipSecondStep = False
         for step in self.__runSteps:
+            
+            if skipNextStep:
+                skipNextStep = False
+                continue
+            if skipSecondStep:
+                skipSecondStep = False
+                skipNextStep   = True
+                
+            # step[0] is the callable step function
+            # step[1] is 'args' positionnal arguments
+            # step[2] is 'kwargs' keywords arguments
             self.logger.debug("run :: step[0]: %s", str(step[0]))
             self.logger.debug("run :: step[1:] %s", str(step[1:]))
+            isBarrierStep=(step[0].__name__==self.barrier.__name__)
+            # Trace the step (not in the barrier step
+            # it will be traced when called ? FIXME is it a good way to do that ?
+            if (not isBarrierStep):
+                self.dtestmaster.traceManager.traceStep(self,self,step)
+            # Step has argument positionnal 'args' and/or keyword 'kwargs' 
             if (len(step[1]) > 0) or (len(step[2]) > 0):
                 try:
-                    if (DTester.__isbound(step[0])):
-                        step[0](*(step[1]),**(step[2]))
+                    if (DTester.__isbound(step[0])):                        
+                        #for pseudo-execution
+                        if ((not self.dtestmaster.pseudoexec)or isBarrierStep):
+                            self.__lastStepStatus = step[0](*(step[1]),**(step[2]))
                     else:
-                        step[0](self,*(step[1]),**(step[2]))
+                        #for pseudo-execution
+                        if ((not self.dtestmaster.pseudoexec) or isBarrierStep):
+                            self.__lastStepStatus = step[0](self,*(step[1]),**(step[2]))
                 except NotImplementedError, err:
                     self.logger.error("RunStep <%r> failed <%r>" % (step[0],err))
-                    self.ok(False,desc="RunStep <%r> failed <%r>" % (step[0],err))
+                    self.__lastStepStatus = False
+                    self.dtestmaster.traceManager.traceStepResult(False,desc="RunStep <%r> failed <%r>" % (step[0],err))
             else:
                 try:
-                    if (DTester.__isbound(step[0])):
-                        step[0]()
-                    else:
-                        step[0](self)
+                    if (DTester.__isbound(step[0])):                        
+                        #for pseudo-execution
+                        if ((not self.dtestmaster.pseudoexec) or isBarrierStep):
+                            self.__lastStepStatus = step[0]()
+                    else:                        
+                        #for pseudo-execution
+                        if ((not self.dtestmaster.pseudoexec) or isBarrierStep):
+                            self.__lastStepStatus = step[0](self)
                 except NotImplementedError, err:
                     self.logger.error("Step <%r> failed <%r>" % (step[0],err))
-                    self.ok(False,desc="Step <%r> failed <%r>" % (step[0],err))
+                    self.__lastStepStatus = False
+                    self.dtestmaster.traceManager.traceStepResult(False,desc="RunStep <%r> failed <%r>" % (step[0],err))
+            # skip next step if a ifThenElse was just executed
+            if (step[0].__name__ == self.ifThenElse.__name__):            
+                if not self.__lastStepStatus:
+                    self.logger.debug("skipping next step");
+                    skipNextStep = True                    
+                else:
+                    skipSecondStep = True
+            self.logger.debug("lastStepStatus = %r" % self.__lastStepStatus)
         self.logger.info("DTester <%s> has terminated." % self.getName())
